from app.setup_http                 import start_http_server
from app.setup_ws                   import start_websocket_server
from app.setup_db                   import start_db_daemon, start_db_client
from app.setup_active_learner       import start_vowpal_wabbit_daemon, start_active_learner, active_learner_with_seeding, active_learner_with_inference, _push_basic_knowledge
from app.setup_target_application   import start_target_application
from app.setup_event_loop           import start_event_loop
from app.setup_app                  import start_sampler
from app.setup_app_configuration    import get_configuration
from app.setup_app_logging          import setup_logging
from app.setup_app_persistence      import setup_persistence

from sampling.generators.fast import get_fast_generator
from sampler import Sampler

from event_dispatching import EventEmitter


from net.netutils import DummyNet


import logging


def _setup_http_server(configuration):
    return start_http_server(configuration['host'], configuration['port'])

def _setup_websocket_server(ee, configuration):
    return start_websocket_server(configuration['host'], configuration['port'], ee)

def _setup_db_daemon(configuration):
    return start_db_daemon(configuration['port'], configuration['data_location'])

def _setup_db_client(configuration):
    return start_db_client(configuration['host'], configuration['port'], configuration['name'], configuration['collections'])

def _setup_vw_daemon(configuration):
    return start_vowpal_wabbit_daemon(configuration['port'], configuration['mellowness'])

def _setup_active_learner(ee, collections, configuration):
    point_generator = get_fast_generator(configuration['space'], continuous = True)
    active_learner = start_active_learner(ee, configuration['host'], configuration['port'], point_generator)
    active_learner = active_learner_with_seeding(active_learner, collections)
    active_learner = active_learner_with_inference(active_learner, collections, ee, thresholds=configuration['thresholds'])

    active_learner.submit_point({
        'download_lossrate': configuration['space']['download_lossrate']['begin'],
        'download_throughput': configuration['space']['download_throughput']['end'],
        'rtt': configuration['space']['rtt']['begin'],
        'label': 0
    })
    active_learner.submit_point({
        'download_lossrate': configuration['space']['download_lossrate']['end'],
        'download_throughput': configuration['space']['download_throughput']['begin'],
        'rtt': configuration['space']['rtt']['end'],
        'label': 1
    })

    return active_learner

def _setup_target_application(ee, server, configuration):
    return start_target_application(ee, server.host, server.port, configuration['user'], configuration['browser_cmd'], configuration['messages'])

if __name__ == "__main__":
    db_server   = None
    vw_server   = None
    http_server = None
    ws_server   = None
    event_loop  = None

    def _terminate(target, name):
        logging.info("[__MAIN__]: closing {}".format(name))
        target.stop()
    #    target.join()
        logging.info("[__MAIN__]: {} closed".format(name))

    # termination routine to close the program (stop all active entities)
    def terminate(signum = None, frame = None):
        global db_server, vw_server, http_server, ws_server, event_loop

        targets = [
            (event_loop,  "event_loop"          ),
            (db_server,   "db_daemon"           ),
            (vw_server,   "vowpal wabbit daemon"),
            (http_server, "http server"         ),
            (ws_server,   "websocket server"    )
    #        (client, "target application")
        ]

        for target in targets:
            try:
                _terminate(target[0], target[1])
            except:
                import traceback
                traceback.print_exc()

    # init core environment
    # # setup logging
    setup_logging()
    # # setup event based framework
    ee = EventEmitter()
    event_loop = start_event_loop(ee)

    # reset DummyNet
    dummynet = DummyNet()
    dummynet.reset_network_condition()
    #TODO dummynet reference should be passed to the executor and tester instance (now everything works because it's a singleton but consider it as a quickfix)

    # load configuration
    configuration = get_configuration()
    try:
        # init app services
        http_server = _setup_http_server(configuration['http'])
        ws_server   = _setup_websocket_server(event_loop, configuration['websocket'])
        db_server   = _setup_db_daemon(configuration['db'])

        # start application
        (conn, db, collections) = _setup_db_client(configuration['db'])
        setup_persistence(ee, collections)

        vw_server       = _setup_vw_daemon(configuration['active_learner'])
        active_learner  = _setup_active_learner(ee, collections, configuration['active_learner'])

        count = 0
        def inject(*args, **kwargs):
            global count, active_learner
            if count == 10:
                _push_basic_knowledge(active_learner)
                count = 0

            count += 1

        ee.on(Sampler.Events.SAMPLE_COLLECTED, inject)

        client          = _setup_target_application(event_loop, http_server, configuration['client'])
        sampler         = start_sampler(event_loop, active_learner, client, configuration['sampler']) #NB: event loop ensures thread safety (being a serialized event emitter)
    except:
        import traceback
        traceback.print_exc()
        terminate()

    logging.info("[__MAIN__]: setup signal handling")
    from signal import signal, SIGINT, pause
    signal(SIGINT, terminate)
    logging.info("[__MAIN__]: press ^C to interrupt")
    pause()
