# network tools (for video streaming)
# import dpkt, pcap # traffic capture tools
# import ipfw # ipfw minimal wrapper to set up DummyNet configurations
from netutils import DummyNet, PcapSniffer, resolve_url_to_ip

# socket support (used for rtt test)
import socket

# threading and external process execution
from threading  import Timer, Thread, Condition #, Semaphore
from subprocess import Popen, PIPE

# utilities for process handling (communication)
from subprocessutils import exec_command

from net.tsharkutils import filter_pcap_by_ip, calculate_losses, calculate_outgoing_throughput, calculate_ack_rtt

# to use the info.json file creted by youtube-dl
import json

# url parser (to remove useless part from video server url in order to perform IP resolution)
from urlparse import urlparse

# import regex support
import re

# EventEmitter - event-driven interface support
from pymitter import EventEmitter

# timing utils
import time

# RTT Tester (using HTTP requests / 3-way handshaking)
#TODO outpput should be an object containing all the details of the sampling (like DL tester)
class RttTester(object):
    STARTED              = "STARTED"
    SNIFFING_STARTED     = "SNIFFING_STARTED"
    HTTP_REQUEST_STARTED = "HTTP_REQUEST_STARTED"
    HTTP_REQUEST_STOPPED = "HTTP_REQUEST_STOPPED"
    SNIFFING_TERMINATED  = "SNIFFING_TERMINATED"
    ANALYSIS_STARTED     = "ANALYSIS_STARTED"
    ANALYSIS_TERMINATED  = "ANALYSIS_TERMINATED"
    TERMINATED           = "TERMINATED"

    """RttTester"""
    def __init__(self, ee = EventEmitter()):
        super(RttTester, self).__init__()

        self.ee = ee

        self._th_pcap = None
        self.sniffing = False

    # BLOCKING FUNCTION
    """
    perform several connections to the 80 port of the target_url (only 3-way handshaking then close) and analyze the captured traffic

    @arg output_path location where the download will be performed (valid path with trailing / )
    @arg output_file_id prefix used to identify all the files generated by the tester
    @arg target_url String target media url
    @arg number_conn_requests int number of connection tests
    @arg conf (optional) Not In Use, meant to contains other options for YDL
    @arg thread_manager (optional)
    """
    #TODO introduce timeout?
    #@arg timeout (optional) timeout after the download is aborted (None to disable it)
    def perform_rtt_test(self, output_path, output_file_id, target_url, number_conn_requests = 100, conf = {}, thread_manager = None, verbose = True):
        self.conf = conf


        if output_path[-1] is not '/':
             output_path = output_path + '/'

        self._file_root = output_path + output_file_id

        self.file_pcap     = self._file_root + ".pcap"
        self.file_pcap_ip  = self._file_root + "-ip.pcap"


        self.target_url           = target_url
        self.number_conn_requests = number_conn_requests

        init_fn  = lambda: self._dump(self.file_pcap)

        self.is_stopped_externally = False
        close_fn = lambda: self.stop()

        # start pcap sniffing
        self.sniffing = True
        self.time_sniffer_started = time.time()
        if thread_manager is not None:
            self._th_pcap = thread_manager.register_thread(target = init_fn, close_fn = close_fn, args = (), name = "pcap-capture-{}-rtt_test".format(output_file_id))
        else:
            self._th_pcap = Thread(target = init_fn, args = (), name = "pcap-capture-{}-rtt_test".format(output_file_id))
        self._th_pcap.start()

        # perform conn requests
        # requester_condition      = Condition() #TODO not in use?
        requests_completed_count = 0
        requests_to_perform      = number_conn_requests

        ## SEQUENTIALLY PERFORM conn REQUESTS
        self.__prev_bin = 0
        if verbose:
            print "sending active probes/"
        while not self.is_stopped_externally and requests_completed_count <= requests_to_perform:
            #urllib2.urlopen(target_url).read()
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((target_url, 80)) # perform 3-way handshake
            time.sleep(0.5)
            s.close()
            time.sleep(0.5)

            if verbose:
                self._update_progress_bar(requests_completed_count, requests_to_perform)

            requests_completed_count += 1
        ##\
        if verbose:
            print "\nsending active probes\\"

        # stop pcap sniffing
        self.stop_sniffing()


        # get video server IP
        target_addr = resolve_url_to_ip(target_url)

        # filter communication between client and server
        filter_pcap_by_ip(self.file_pcap, self.file_pcap_ip, target_addr, self.time_sniffer_started) # , self.time_sniffer_stopped omitted

        # calculate rtt from pcap file (during the transfer time)
        rtt_list = calculate_ack_rtt(
            self.file_pcap_ip,
            target_addr,
            (self.time_sniffer_stopped - self.time_sniffer_started) # transfer time
        )

        rtt = rtt_list[0] # single value enforced by the transfer time


        # calculate packet losses in the whole transmission period
        losses_list = calculate_losses(
            self.file_pcap_ip,
            target_addr,
            (self.time_sniffer_stopped - self.time_sniffer_started) # transfer time
        )

        res = {}

        res['rtt'] = rtt

        res['packet_count'] = {}
        res['packet_count']['download']      = losses_list[0]['src_transmitted']
        res['packet_count']['upload']        = losses_list[0]['dst_transmitted']
        res['packet_count']['all']              = res['packet_count']['download'] + res['packet_count']['upload']
        res['retransmissions'] = {}
        res['retransmissions']['download']      = losses_list[0]['src_retransmissions']
        res['retransmissions']['upload']        = losses_list[0]['dst_retransmissions']
        res['duplicate_acks'] = {}
        res['duplicate_acks']['download']       = losses_list[0]['src_duplicate_acks']
        res['duplicate_acks']['upload']         = losses_list[0]['dst_duplicate_acks']
        res['lost_segments'] = {}
        res['lost_segments']['download']        = losses_list[0]['src_lost_segments']
        res['lost_segments']['upload']          = losses_list[0]['dst_lost_segments']
        res['fast_retransmissions'] = {}
        res['fast_retransmissions']['download'] = losses_list[0]['src_fast_retransmissions']
        res['fast_retransmissions']['upload']   = losses_list[0]['dst_fast_retransmissions']

        res['experiment'] = {}
        res['experiment']['conf']                  = self.conf
        res['experiment']['target'] = {}
        res['experiment']['target']['url']         = self.target_url
        res['experiment']['is_stopped_externally'] = self.is_stopped_externally
        res['experiment']['server'] = {}
        res['experiment']['server']['url']         = target_url
        res['experiment']['server']['ip']          = target_addr

        res['output'] = {}
        res['output']['file_pcap']                 = self.file_pcap
        res['output']['file_pcap_ip']              = self.file_pcap_ip

        return res

    def _update_progress_bar(self, current, total):
        current_bin = int((float(current) / total) * 100)

        if self.__prev_bin < current_bin:
            delta_bin = current_bin - self.__prev_bin

            import sys
            sys.stdout.write('|' * delta_bin)
            sys.stdout.flush()

            self.__prev_bin = current_bin


    def _dump(self, output_filename):
        self._sniffer = PcapSniffer()
        self._sniffer.dump(output_filename)

    #TODO should not be public
    # useful to keep separate from stop because this happens also during the normal behaviour
    def stop_sniffing(self):
        if self._sniffer is not None:
            self._sniffer.close()
            self.time_sniffer_stopped = time.time()
            self._sniffer = None
            self._th_pcap = None

            self.sniffing = False

    def stop(self):
        self._set_stopped_externally()
        self._stop()

    #TODO prevent to continue tests if experiment aborted
    def _stop(self):
        # stop pcap sniffing
        self.stop_sniffing()

    def _set_stopped_externally(self):
        self.is_stopped_externally = True


#TODO rename in MediaDlTester
#  Video Streaming Tester
class VideoDlTester(object):
    STARTED             = "STARTED"
    SNIFFING_STARTED    = "SNIFFING_STARTED"
    VIDEO_REQUESTED     = "VIDEO_REQUESTED"
    VIDEO_DOWNLOADING   = "VIDEO_DOWNLOADING"
    VIDEO_DOWNLOADED    = "VIDEO_DOWNLOADED"
    SNIFFING_TERMINATED = "SNIFFING_TERMINATED"
    ANALYSIS_STARTED    = "ANALYSIS_STARTED"
    ANALYSIS_TERMINATED = "ANALYSIS_TERMINATED"
    TERMINATED          = "TERMINATED"

    EVENTS = [
        STARTED,
        SNIFFING_STARTED,
        VIDEO_REQUESTED,
        VIDEO_DOWNLOADING,
        VIDEO_DOWNLOADED,
        SNIFFING_TERMINATED,
        ANALYSIS_STARTED,
        ANALYSIS_TERMINATED,
        TERMINATED
    ]

    """VideoDlTester"""
    def __init__(self, ee = EventEmitter()):
        super(VideoDlTester, self).__init__()

        self.ee = ee

        self._ydl_process = None
        self._th_time     = None
        self._th_pcap     = None
        self.sniffing     = False


    # BLOCKING FUNCTION
    # conf must contains: target_url, target_video_quality (not the necessarly dummynet_conf)
    #TODO consider what else introduce in conf
    """
    perform a video download (ftp like) and analyze the captured traffic

    @arg output_path location where the download will be performed (valid path with trailing / )
    @arg output_file_id prefix used to identify all the files generated by the tester
    @arg target_url String target media url
    @arg target_quality String target media quality
    @arg conf (optional) Not In Use, meant to contains other options for YDL
    @arg timeout (optional) timeout after the download is aborted (None to disable it)
    @arg thread_manager (optional)
    """
    def perform_dl_test(self, output_path, output_file_id, target_url, target_quality = 'best', conf = {}, timeout = None, thread_manager = None):
        self.target_url     = target_url
        self.target_quality = target_quality
        self.conf           = conf

        self._ydl = None

        if output_path[-1] is not '/':
             output_path = output_path + '/'

        self._file_root = output_path + output_file_id

        self.file_video    = self._file_root + ".video.dat"
        self.file_ydl_info = self._file_root + ".video.info.json"
        self.file_pcap     = self._file_root + ".pcap"

        self.file_pcap_ip  = self._file_root + "-ip.pcap"

        init_fn    = lambda: self._dump(self.file_pcap)
        close_fn   = lambda: self._set_stopped_externally(), self.stop_ydl()
#        timeout_fn = lambda: self._set_stopped_by_timeout(), self.stop_ydl() # FAILS in Timer! It's seen as a tuple (not callables)

        def timeout_fn():
            self._set_stopped_by_timeout()
            self.stop_ydl()

        #TODO use timed context to close timer ASAP and avoid interferences in multiple call scenarios
        # setup timeout thread (abort experiment after some time)
        self.is_stopped_by_timeout = False
        if timeout is not None:
            self._th_time = Timer(timeout, timeout_fn)
            self._th_time.setName('dl_test-timer')
            self._th_time.start()

        # start pcap sniffing
        self.sniffing = True
        self.time_sniffer_started = time.time()
        self.is_stopped_externally = False
        if thread_manager is not None:
            self._th_pcap = thread_manager.register_thread(target = init_fn, close_fn = close_fn, args = (), name = "pcap-capture-{}-dl_test".format(output_file_id))
        else:
            self._th_pcap = Thread(target = init_fn, args = (), name = "pcap-capture-{}-dl_test".format(output_file_id))
        self._th_pcap.start()

        # start ydl process handling (blocking)
        ydl_exec_info = self._handle_ydl_process(self.file_video)

        # stop pcap sniffing
        self.stop_sniffing()

        # get video server URL
        with open(self.file_ydl_info) as ydl_info_file:
            # load ydl info file
            ydl_info = json.load(ydl_info_file)

            # get url (hostname address only)
            target_url = urlparse(ydl_info['url']).netloc

        # get video server IP
        target_addr = resolve_url_to_ip(target_url)

        # update ydl info file
        ydl_info['ip'] = target_addr

        # filter pcap file by server IP
        filter_pcap_by_ip(self.file_pcap, self.file_pcap_ip, target_addr, ydl_exec_info['dl_time']['start']) # , ydl_exec_info['dl_time']['end'] omitted

        # calculate throughput for each second
        throughput_list = calculate_outgoing_throughput(self.file_pcap_ip, target_addr, 1) # rename into data_volume_list?

        #TODO do it here?
        # discard first and last value (not representative because it could be truncated)
        #TODO only the last one can be truncated?
        throughput_list = throughput_list[1:-1]


        # calculate packet losses in the whole transmission period
        losses_list = calculate_losses(
            self.file_pcap_ip,
            target_addr,
            (self.time_sniffer_stopped - self.time_sniffer_started)
        )

        # define result
        #TODO return tuple with all data captured during the experiment?
        # throughput, packet_loss, info_file, target_addr, target_ip, timing

        res = {}

        res['throughput'] = {}
        res['throughput']['download']           = throughput_list

        res['packet_count'] = {}
        res['packet_count']['download']         = losses_list[0]['src_transmitted']
        res['packet_count']['upload']           = losses_list[0]['dst_transmitted']
        res['packet_count']['all']              = res['packet_count']['download'] + res['packet_count']['upload']
        res['retransmissions'] = {}
        res['retransmissions']['download']      = losses_list[0]['src_retransmissions']
        res['retransmissions']['upload']        = losses_list[0]['dst_retransmissions']
        res['duplicate_acks'] = {}
        res['duplicate_acks']['download']       = losses_list[0]['src_duplicate_acks']
        res['duplicate_acks']['upload']         = losses_list[0]['dst_duplicate_acks']
        res['lost_segments'] = {}
        res['lost_segments']['download']        = losses_list[0]['src_lost_segments']
        res['lost_segments']['upload']          = losses_list[0]['dst_lost_segments']
        res['fast_retransmissions'] = {}
        res['fast_retransmissions']['download'] = losses_list[0]['src_fast_retransmissions']
        res['fast_retransmissions']['upload']   = losses_list[0]['dst_fast_retransmissions']

        res['experiment'] = {}
        res['experiment']['conf']                  = self.conf
        res['experiment']['target'] = {}
        res['experiment']['target']['url']         = self.target_url
        res['experiment']['target']['quality']     = self.target_quality
        res['experiment']['is_stopped_by_timeout'] = self.is_stopped_by_timeout
        res['experiment']['is_stopped_externally'] = self.is_stopped_externally
        res['experiment']['server'] = {}
        res['experiment']['server']['url']         = target_url
        res['experiment']['server']['ip']          = target_addr

        res['output'] = {}
        res['output']['file_video']                = self.file_video
        res['output']['file_ydl_info']             = self.file_ydl_info
        res['output']['file_pcap']                 = self.file_pcap
        res['output']['file_pcap_ip']              = self.file_pcap_ip

        # update json file
        with open(self.file_ydl_info, 'w') as fp:
            json.dump(ydl_info, fp)


        return res

    def _set_stopped_by_timeout(self):
        self.is_stopped_by_timeout = True

    def _set_stopped_externally(self):
        self.is_stopped_externally = True

    def _handle_ydl_process(self, output_filename):
        self._ydl = YoutubeDl()
        return self._ydl.execute(
            self.target_url,
            self.target_quality,
            output_filename)

    def _dump(self, output_filename):
        self._sniffer = PcapSniffer()
        self._sniffer.dump(output_filename)

    def cancel_timeout(self):
        if self._th_time is not None:
            self._th_time.cancel() #TODO check on documentation
            self._th_time = None

    #TODO rename in stop_media_downloader
    def stop_ydl(self):
        if self._ydl is not None:
            self._ydl.stop()
            self._ydl = None

    def stop_sniffing(self):
        if self._sniffer is not None:
            self._sniffer.close()
            self.time_sniffer_stopped = time.time()
            self._sniffer = None
            self._th_pcap = None

            self.sniffing = False

    # blocking function (stops all the executing thread and abort the experiment if it's running)
    def stop(self):
        # abort timer (if active)
        self.cancel_timeout()

        # kill youtube-dl subprocess
        self.stop_ydl()

        # stop pcap sniffing
        self.stop_sniffing()



# SUPPORTED FORMATS (video) - Youtube
    # nondash-160 = 144p mp4
    # nondash-242 = 240p (small) webm
    # nondash-133 = 240p (small) mp4
    # nondash-243 = 360p (medium) webm
    # nondash-134 = 360p (medium) mp4
    # nondash-244 = 480p (large) webm
    # nondash-135 = 480p (large) mp4
    # nondash-247 = 720p webm
    # nondash-136 = 720p mp4
    # nondash-248 = 1080p webm
    # nondash-137 = 1080p mp4
    # nondash-271 = 1440p webm
    # nondash-264 = 1440p mp4
    # nondash-272 = 2160p webm
    # nondash-313 = 2160p webm VP9
    # nondash-266 = 2160p h264
    # 278 = 144p webm VP9 12fps
    # 160 = 144p mp4 12fps
    # 242 = 240p (small) webm 24fps
    # 133 = 240p (small) mp4 24fps
    # 243 = 360p (medium) webm 24fps
    # 134 = 360p (medium) mp4 24fps
    # 244 = 480p (large) webm 24fps
    # 247 = 720p webm 24fps
    # 136 = 720p mp4 24fps
    # 248 = 1080p webm 24fps
    # 137 = 1080p mp4 24fps
    # 271 = 1440p webm 24fps
    # 264 = 1440p mp4 24fps
    # 313 = 2160p webm VP9 24fps
    # 272 = 2160p webm 24fps
    # 266 = 2160p mp4 h264 24fps
    # 138 = 2160p mp4 24fps
    # 17  = 144p 3gp
    # 36  = 240p (small) 3gp
    # 5   = 240p (small) flv
    # 43  = 360p (medium) webm
    # 18  = 360p (medium) mp4
    # 22  = 720p mp4

class YoutubeDl(object):
    STARTED           = "STARTED"
    VIDEO_REQ_CMD_SET = "VIDEO_REQ_CMD_SET" # video request command set
    VIDEO_REQUESTED   = "VIDEO_REQUESTED"
    VIDEO_DOWNLOADING = "VIDEO_DOWNLOADING"
    VIDEO_DOWNLOADED  = "VIDEO_DOWNLOADED"
    TERMINATED        = "TERMINATED"

    EVENTS = [
        STARTED,
        VIDEO_REQ_CMD_SET,
        VIDEO_REQUESTED,
        VIDEO_DOWNLOADING,
        VIDEO_DOWNLOADED,
        TERMINATED
    ]

    """YoutubeDl"""
    def __init__(self, ee = EventEmitter()):
        super(YoutubeDl, self).__init__()

        self.ee = ee

        self._ydl_process = None

    def on(self, evt, handler):
        if evt not in YoutubeDl.EVENTS:
            print("Warning: {} is not raised by the YoutubeDl class (handler registered anyway)")

        self._ee.on(evt, handler)

    def execute(self, target_url, target_video_quality, output_filename): #TODO add args to support functions of youtube-dl
        self.ee.emit(YoutubeDl.STARTED, source = self) #TODO define args (e.g. add parameters to the command)

        output_filename_root = output_filename.rsplit('.', 1)
        output_filename_root = output_filename_root[0]

        # set flags
        self._interrupted = False # download interrupted (stop method called)

        self.cmd = "youtube-dl \"{0}\" --no-cache-dir --no-continue --no-part --write-info-json --prefer-insecure -f {1} -o \"{2}\"".format(
            target_url,
            target_video_quality,
            output_filename
        )
        print self.cmd

        self.ee.emit(YoutubeDl.VIDEO_REQ_CMD_SET, source = self, cmd = self.cmd) #TODO support options

        # init youtube-dl process
        self._ydl_process = Popen(self.cmd, shell = True, stdout = PIPE)

        self.ee.emit(YoutubeDl.VIDEO_REQUESTED, source = self, target_url = target_url, target_video_quality = target_video_quality)

        # get video download start time
        self._begin_download = None
        regex = re.compile("\[download\] Destination: *")
        for line in iter(self._ydl_process.stdout.readline, ''):
            if(regex.match(line)):
                self._begin_download = time.time()
                break

        if self._interrupted:
            raise RuntimeError("youtube-dl interrupted externally")

        if self._begin_download is None:
            raise RuntimeError("youtube-dl failed it's execution - no download started")

        self.ee.emit(YoutubeDl.VIDEO_DOWNLOADING, source = self, time = self._begin_download)

        # execute the whole process (until it finishes or it's aborted)
        status_code = self._ydl_process.wait()

        # get video download end time
        self._end_download = time.time()

        # check if the process returned with errors
        if(status_code == 2):
            # self.log("ERROR: ydl process returned failed")
            # os._exit(2) # force exit
            raise RuntimeError("youtube-dl failed it's execution - returned with status code 2")


        self.ee.emit(YoutubeDl.VIDEO_DOWNLOADED, source = self, time = self._end_download, interrupted = self._interrupted) #TODO file to destroy - here or out? (probably out)

        # stop process
        self._stop()

        res = {
            'dl_time': {
                'start': self._begin_download,          # download start time
                'end':   self._end_download             # download end time
            },
            'dl_interrupted': self._interrupted,        # video downloaded interrupted flag
            'cmd': self.cmd,                            # executed command
            'json': output_filename_root + '.info.json' # json file location
        }

        self.ee.emit(YoutubeDl.TERMINATED, source = self, result = res)

        return res

    def is_alive(self):
        return self._ydl_process is not None and self._ydl_process.poll() is not None

    def _stop(self):
        # if process is defined and running
        if self.is_alive():
            try:
                # abort process
                self._ydl_process.kill()
            except Exception:
                pass

        self._ydl_process = None
        self.cmd = None

    def stop(self):
        # if process is defined and running
        if self.is_alive():
            self._interrupted = True

        self._stop()






# class YoutubeUrlResolver(object):
#     """docstring for YoutubeUrlResolver"""
#     def __init__(self, ee = EventEmitter()):
#         super(YoutubeUrlResolver, self).__init__()
#         self.ee = ee

#     def solve_url(url):
#         cmd = 'youtube-dl --skip-download -g {}'.format(url)

#         return exec_command(cmd, is_single_result = True)

def solve_url(url):
    cmd = 'youtube-dl --skip-download -g {}'.format(url)

    return exec_command(cmd, is_single_result = True)
