# network utilities

# network tools
import          dpkt, pcap  # traffic capture tools
from net import ipfw        # ipfw minimal wrapper to set up DummyNet configurations

# threading and external process execution
from threading  import Timer, Thread, Semaphore

# url parser (for url to ip resolution)
from urlparse import urlparse

# utilities for process handling (communication)
from subprocessutils import exec_command

# exception tools
import traceback

class PcapSniffer(object):
    """PcapSniffer"""
    def __init__(self, pcap = pcap.pcap()):
        super(PcapSniffer, self).__init__()
        
        self.pcap = pcap

        self.sniffing   = False
        self._semaphore = Semaphore(1)

    # keep on sniffing until self.sniffing is not False
    def dump(self, output_filename, output_mode = 'w'):
        self.sniffing = True

        # begin critical section
        self._semaphore.acquire()

        # TODO support also fp instead of file name?
        # if isinstance(output_filename, basestring):
        #     output_filename = open(output_filename, output_mode)

        with open(output_filename, output_mode) as out_fp:
            # with clause is not supported by dpkt.pcap.Writer
            # with dpkt.pcap.Writer(out_fp) as out_pcap:

            out_pcap = dpkt.pcap.Writer(out_fp)

            while self.sniffing:
                res = self.pcap.next()
                if res is not None:
                    ts, pkt = res
                    out_pcap.writepkt(pkt)
           
            out_pcap.close()

        # end critical section
        self._semaphore.release()

    def close(self):    
        self.sniffing = False

        # wait until the sniffing operation is completed
        self._semaphore.acquire()

        # sniffing completed
        self._semaphore.release()

'''
Wrapper for PcapSniffer in order to use it with the python construct 'with' (define sniffing sections)
'''
#TODO support deamon argument for thread
class PcapSniffingEnvironment(object):
    def __init__(self, output_filename, output_mode = 'w', pcap = pcap.pcap(timeout_ms = 500), thread_manager = None, name = 'pcap_sniffer'):
        self.output_filename = output_filename
        self.output_mode     = output_mode

        self.thread_manager  = thread_manager
        self.name            = name

        self._sniffer        = PcapSniffer(pcap)

    def __enter__(self):
        self.start()
        return self

    def start(self):
        #TODO put in __init__?

        if self.thread_manager is not None:
            self._sniffer_th = self.thread_manager.register_thread(target = lambda: self._sniffer.dump(self.output_filename, self.output_mode), close_fn = self.stop, name = self.name)
        else:
            self._sniffer_th = Thread(target = lambda: self._sniffer.dump(self.output_filename, self.output_mode), name = self.name)
        
        self._sniffer_th.start()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            #print exc_type, exc_value, traceback # display exception (if any)
            print traceback.format_exc()

        self.stop()
        
        return True # suppress exceptions (if any)

    def stop(self):
        self._sniffer.close()


# dummynet wrapper (of the wrapper...)
class DummyNet(object):
    NOT_CONSTRAINED = {
        'download': {
            'lossrate':  0, # from 0 to 1 (100%)
            'bandwidth': 0, # in Kbps
            'delay':     0  # in ms
        },
        'upload': {
            'lossrate':  0, # from 0 to 1 (100%)
            'bandwidth': 0, # in Kbps
            'delay':     0  # in ms
        }
    }

    """DummyNet"""
    def __init__(self):
        super(DummyNet, self).__init__()
        
    def set_network_condition(self, conf = NOT_CONSTRAINED, pipe_id = None):
        if conf != DummyNet.NOT_CONSTRAINED:
            print "DummyNet set configuration"

        if pipe_id is None:
            # use default ones (10 for upload, 11 for download)
            ipfw.configDownloadPipe(conf["download"]["bandwidth"], conf["download"]["delay"], conf["download"]["lossrate"])
            ipfw.configUploadPipe(conf["upload"]["bandwidth"], conf["upload"]["delay"], conf["upload"]["lossrate"])
        else:
            ipfw.configDownloadPipe(pipe_id, conf["download"]["bandwidth"], conf["download"]["delay"], conf["download"]["lossrate"])
            ipfw.configUploadPipe(pipe_id, conf["upload"]["bandwidth"], conf["upload"]["delay"], conf["upload"]["lossrate"])

    def reset_network_condition(self, pipe_id = None):
        print "DummyNet reset"
        self.set_network_condition(pipe_id = pipe_id)

'''
Wrapper for DummyNet in order to use it with the python construct 'with' (define sections where the network is controlled via DummyNet)
'''
class DummyNetEnvironment(object):
    def __init__(self, conf, pipe_id = None):
        self.conf    = conf
        self.pipe_id = pipe_id

    def __enter__(self):
        self.start()
        return self

    def start(self):
        self._dummynet = DummyNet()
        self._dummynet.set_network_condition(self.conf, self.pipe_id)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            #print exc_type, exc_value, traceback # display exception (if any)
            print traceback.format_exc()

        self.stop()

        return True # suppress exceptions (if any)

    def stop(self):
        self._dummynet.reset_network_condition(self.pipe_id)

'''
Resolve an url to it's ip address
'''
def resolve_url_to_ip(target_url):
    url = urlparse(target_url)
    target_url_netloc = url.netloc if url.netloc is not '' else url.path # take first part of url (the one that identifies an host and not a resource or some request parameters)

    # host supports only url with the network location part (see step before to ensure this)
    _cmd = "host \"{0}\" | grep -Eo \"[0-9]{{1,3}}\\.[0-9]{{1,3}}\\.[0-9]{{1,3}}\\.[0-9]{{1,3}}+\"".format(target_url_netloc)
    return exec_command(_cmd, None, True) # , verbose_mode = True

